'''Some definitions for the snom_analysis package. Mainly enums for internal referencing.'''

##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

from enum import Enum, auto

# keep this for internal referencing
class Definitions(Enum):
    """This class keeps track of the implemented definitions."""
    vertical = auto()
    horizontal = auto()

class MeasurementTypes(Enum):
    AFM = auto()
    SNOM = auto()
    APPROACHCURVE = auto()
    SCAN3D = auto()
    SPECTRUM = auto()
    NONE = auto()

class MeasurementTags(Enum):
    """This class keeps track of the implemented measurement tags. 
    These are tags wich are measurement specific.
    Only tags which are listed here can be used.
    """
    SCAN = auto()   # scan type, afm, snom, approach curve, 2d/3d, PsHet...
    PROJECT = auto()   
    DESCRIPTION = auto()   
    DATE = auto()   
    SCANNERCENTERPOSITION = auto()   
    ROTATION = auto()   
    SCANAREA = auto()  
    PIXELAREA = auto()   
    AVERAGING = auto()   
    INTEGRATIONTIME = auto()   
    LASERSOURCE = auto()   
    DETECTOR = auto()   
    TARGETWAVELENGTH = auto()    
    DEMODULATIONMODE = auto()   
    TIPFREQUENCY = auto()   
    TIPAMPLITUTDE = auto()   
    TAPPINGAMPLITUDE = auto()   
    MODULATIONFREQUENCY = auto()   
    MODULATIONAMPLITUDE = auto()   
    MODULATIONOFFSET = auto()   
    SETPOINT = auto()   
    REGULATOR = auto()   
    TIPPOTENTIAL = auto()   
    M1ASCALING = auto()  
    QFACTOR = auto()   
    VERSION = auto()   

class ChannelTags(Enum):
    """This class keeps track of the implemented channel tags.
    These are tags which are channel specific.
    So multiple channels might have varying channel tag values but
    they will share the same measurement tag values. However, the 
    current channel tag values are always to prefer over the same measurement tag value.
    As the channel tag values change when the channel is manipulated.
    """
    PIXELAREA = auto()  
    YINCOMPLETE = auto() 
    SCANNERCENTERPOSITION = auto()   
    ROTATION = auto()   
    SCANAREA = auto()  
    XYUNIT = auto()
    ZUNIT = auto()
    WAVENUMBERSCALING = auto()
    # additional tags
    PIXELSCALING = auto()
    # additional for aachen files (.dump)
    # INTEGRATIONTIME = auto()
    # TIPFREQUENCY = auto()
    # MODULATIONFREQUENCY = auto()
    # TAPPINGAMPLITUDE = auto()
    # MODULATIONOFFSET = auto()
    # SETPOINT = auto()
    
class PlotDefinitions:
    """This class contains all the definitions for the plotting parameters.
    """
    hide_ticks = True
    figsizex = 10
    figsizey = 5
    show_titles = True
    tight_layout = True
    colorbar_width = 2 # in percent of the fig width, standard is 2
    hspace = 0.4 #standard is 0.4
    # Define Plot font sizes
    font_size_default = 8
    font_size_axes_title = 12
    font_size_axes_label = 10
    font_size_tick_labels = 8
    font_size_legend = 8
    font_size_fig_title = 12
    #definitions for color bar ranges:
    # using the same range for all channels is useful for comparison
    # make all height channels have the same range?
    height_cbar_range = False
    vmin_height = None
    vmax_height = None
    # make all amplitude channels have the same range?
    amp_cbar_range = False
    vmin_amp = None#1 # to make shure that the values will be initialized with the first plotting command
    vmax_amp = None#-1
    # phase_cbar_range = True
    # plot the full 2pi range for the phase channels no matter what the actual data range is?
    full_phase_range = True # this will overwrite the cbar
    # make all phase channels have the same range?
    shared_phase_range = False # only used if full phase range is false
    vmin_phase = None
    vmax_phase = None
    real_cbar_range = True
    vlimit_real = None
    # vmin_real = None
    # vmax_real = None
    # show plot automatically? turn to false for gui programming
    show_plot = True
    autodelete_all_subplots = True # if true old subplots will be deleted on creation of new measurement
    # matplotlib style file
    use_mplstyle = False
   


##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import numpy as np

def flatten_phase_profile(profile:list, phase_orientation:int=1) -> list:
    """flattened_profile = []
    previous_element = profile[0]
    offset = 0
    for element in profile:
        if (phase_orientation == 1) and (element - previous_element < -np.pi):
            offset += 2*np.pi
            # print('increased offset')
        elif (phase_orientation == -1) and (element - previous_element > np.pi):
            offset -= 2*np.pi
            # print('reduced offset')
        flattened_profile.append(element + offset)
        previous_element = element

    return flattened_profile"""
    # new alternative, use built-in numpy function unwrap
    return np.unwrap(profile)


def get_smallest_difference(value1, value2):
    # make shure value1 is smaller than value2
    orientation = 1
    if value1 > value2:
        copy = value1
        value1 = value2
        value2 = copy
        # orientation = -1
    difference = abs(value1 - value2)
    if difference > np.pi:
        difference = np.pi*2 - difference
    return difference*orientation

def get_difference(value1, value2):
    difference = value1-value2
    if difference < 0:
        difference += 2*np.pi
    return difference

def get_difference_2(value1, value2):
    return value1-value2
    
def get_profile_difference(profile1:list, profile2:list) -> list:
    # difference = [abs(profile1[i] - profile2[i]) for i in range(len(profile1))]
    # difference = [profile1[i] - profile2[i] for i in range(len(profile1))]
    # difference = [abs(profile1[i] - profile2[i]) if abs(profile1[i] - profile2[i])< np.pi else 2*np.pi - abs(profile1[i] - profile2[i]) for i in range(len(profile1))]
    difference = []
    for i in range(len(profile1)):
        # difference.append(get_smallest_difference(profile1[i], profile2[i]))
        difference.append(get_difference(profile1[i], profile2[i]))

    return difference
    # pass

def get_profile_difference_2(profile1:list, profile2:list) -> list:
    difference = []
    for i in range(len(profile1)):
        difference.append(get_difference_2(profile1[i], profile2[i]))
    return difference

def get_modeindex_from_linearfunction(slope, pixelsize, wavelength=1600):
    # wavelength and pixelsize in nm
    period = np.pi*2/slope*pixelsize
    mode_index = wavelength/period
    return mode_index



##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Button, Slider
from matplotlib.backend_bases import MouseButton
import matplotlib.lines as mlines
# import tkinter as tk
from .snom_colormaps import SNOM_amplitude, SNOM_phase, SNOM_height

import skimage as ski

def select_profile(data, channel):
    # root = tk.Tk()
    # selector = ArraySelector(root, data, channel)
    selector = CutlineSelector(data, channel)
    plt.show()
    # root.mainloop()

    # create profile 
    profile = ski.measure.profile_line(data.T, selector.start, selector.end, linewidth=selector.width) # somehow x and y are switched, therefore transpose the array
    # also return the physical length of the profile and the integration width
    return profile, selector.start, selector.end, selector.width

class CutlineSelector:
    def __init__(self, img_array, channel):
        self.img_array = img_array
        self.channel = channel
        self.fig, self.ax = plt.subplots()
        # determine the colormap
        if ('Z' in self.channel) or ('MT' in self.channel):
            self.cmap = SNOM_height
        elif ('P' or 'arg') in self.channel:
            self.cmap = SNOM_phase
        elif ('A' or 'abs') in self.channel:
            self.cmap = SNOM_amplitude
        elif ('H' or 'height') in self.channel:
            self.cmap = SNOM_height
        else:
            self.cmap = 'gray'
            print('Unknown channel, could not find the proper colormap!')
        self.ax.imshow(img_array, cmap='gray')
        
        # Calculate initial width: 1/20th of the smallest dimension but at least 1
        min_dimension = min(img_array.shape)
        self.width = max(1, min_dimension // 20)

        self.start = None
        self.end = None
        self.dragging_start = False
        self.dragging_end = False
        self.dragging = False
        self.shift_pressed = False
        self.tolerance = 10  # Tolerance in pixels for grabbing endpoints

        # Matplotlib objects for the cutline and boundary lines
        self.main_line, = self.ax.plot([], [], 'r-', linewidth=1.5)
        self.perp_lines = []

        # Connect Matplotlib event handlers
        self.cid_click = self.fig.canvas.mpl_connect('button_press_event', self.on_click)
        self.cid_release = self.fig.canvas.mpl_connect('button_release_event', self.on_release)
        self.cid_motion = self.fig.canvas.mpl_connect('motion_notify_event', self.on_motion)
        self.cid_key_press = self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.cid_key_release = self.fig.canvas.mpl_connect('key_release_event', self.on_key_release)

        # Adding buttons and sliders
        self.add_widgets()

    def add_widgets(self):
        # Adjust button placement
        button_ax = plt.axes([0.82, 0.025, 0.15, 0.04])
        self.button = Button(button_ax, 'Extract Cutline')
        self.button.on_clicked(self.extract_cutline)

        # Slider for adjusting width
        slider_ax = plt.axes([0.15, 0.025, 0.5, 0.04], facecolor='lightgoldenrodyellow')
        self.slider = Slider(slider_ax, 'Width', 1, 100, valinit=self.width, valstep=1)
        self.slider.on_changed(self.update_width)

    def update_width(self, val):
        self.width = int(val)
        self.update_cutline()

    def on_key_press(self, event):
        if event.key == 'shift':
            self.shift_pressed = True

    def on_key_release(self, event):
        if event.key == 'shift':
            self.shift_pressed = False

    def on_click(self, event):
        if event.button == MouseButton.LEFT and event.inaxes == self.ax:
            x, y = event.xdata, event.ydata

            # Check if clicking near the start or end point to enable dragging
            if self.start and self.is_near_point((x, y), self.start):
                self.dragging_start = True
            elif self.end and self.is_near_point((x, y), self.end):
                self.dragging_end = True
            else:
                # Start a new cutline
                self.start = (x, y)
                self.end = None
                self.dragging = True
                self.clear_cutline()

    def on_release(self, event):
        self.dragging = False
        self.dragging_start = False
        self.dragging_end = False

    def on_motion(self, event):
        if event.inaxes != self.ax:
            return

        x, y = event.xdata, event.ydata

        # Change cursor when near endpoints
        if self.start and self.is_near_point((x, y), self.start):
            self.fig.canvas.set_cursor(3)  # Resize cursor
        elif self.end and self.is_near_point((x, y), self.end):
            self.fig.canvas.set_cursor(3)  # Resize cursor
        else:
            self.fig.canvas.set_cursor(1)  # Default cursor

        if self.dragging:
            # Draw the cutline in real-time
            if self.start:
                if self.shift_pressed:
                    x0, y0 = self.start
                    # Snap to horizontal or vertical if shift is pressed
                    if abs(x - x0) > abs(y - y0):
                        y = y0
                    else:
                        x = x0
                self.end = self.clip_point_to_image(x, y)
        elif self.dragging_start:
            # Adjust the start point while keeping orientation
            self.start = self.adjust_endpoint(self.end, (x, y))
        elif self.dragging_end:
            # Adjust the end point while keeping orientation
            self.end = self.adjust_endpoint(self.start, (x, y))

        self.update_cutline()

    def is_near_point(self, p1, p2):
        """Check if point p1 is near point p2 within a certain tolerance."""
        return np.hypot(p1[0] - p2[0], p1[1] - p2[1]) < self.tolerance

    def adjust_endpoint(self, fixed_point, new_point):
        """Adjust endpoint along the fixed orientation."""
        fx, fy = fixed_point
        dx = new_point[0] - fx
        dy = new_point[1] - fy
        direction = np.array([dx, dy])
        unit_direction = direction / np.hypot(*direction)
        length = np.hypot(dx, dy)
        return (fx + unit_direction[0] * length, fy + unit_direction[1] * length)

    def clip_point_to_image(self, x, y):
        """Ensure point (x, y) is within the image bounds."""
        x = np.clip(x, 0, self.img_array.shape[1] - 1)
        y = np.clip(y, 0, self.img_array.shape[0] - 1)
        return x, y

    def clear_cutline(self):
        self.main_line.set_data([], [])
        for pline in self.perp_lines:
            pline.remove()
        self.perp_lines = []
        self.fig.canvas.draw()

    def update_cutline(self):
        if self.start is None or self.end is None:
            return

        # Update the main cutline
        x0, y0 = self.start
        x1, y1 = self.end
        self.main_line.set_data([x0, x1], [y0, y1])

        # Draw perpendicular lines to indicate width
        self.draw_perpendicular_lines(x0, y0, x1, y1)
        self.fig.canvas.draw()

    def draw_perpendicular_lines(self, x0, y0, x1, y1):
        for pline in self.perp_lines:
            pline.remove()
        self.perp_lines = []

        dx, dy = x1 - x0, y1 - y0
        length = np.hypot(dx, dy)
        if length == 0:
            return
        # self.width = length # to be able to access the width of the profile line from outside the class

        # Normalize direction
        dx /= length
        dy /= length

        offset_x = -dy * self.width / 2
        offset_y = dx * self.width / 2

        perp_line1 = mlines.Line2D([x0 + offset_x, x0 - offset_x], [y0 + offset_y, y0 - offset_y], color='r')
        perp_line2 = mlines.Line2D([x1 + offset_x, x1 - offset_x], [y1 + offset_y, y1 - offset_y], color='r')

        self.ax.add_line(perp_line1)
        self.ax.add_line(perp_line2)
        self.perp_lines = [perp_line1, perp_line2]

    def extract_cutline(self, event):
        if self.start is None or self.end is None:
            print("No cutline selected.")
            return
        print(f"Cutline from {self.start} to {self.end} extracted.")
        # close the figure
        plt.close(self.fig)



##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
# import matplotlib as mpl
# to use perceptual colormaps:
import colorcet as cc
"""
SNOM_height = 'gray'

cmap_snom_amplitude = {'red':   ((0.0, 0.0, 0.0),
           (0.33, 1.0, 1.0),
           (0.66, 1.0, 1.0),
           (1.0, 1.0, 1.0)),

  'green': ((0.0, 0.0, 0.0),
           (0.33, 0.0, 0.0),
           (0.66, 1.0, 1.0),
           (1.0, 1.0, 1.0)),

  'blue':  ((0.0, 0.0, 0.0),
           (0.33, 0.0, 0.1),
           (0.66, 0.0, 0.0),
           (1.0, 0.0, 1.0)),

  'alpha': ((0.0, 1.0, 1.0),
            (0.33, 1.0, 1.0),
            (0.66, 1.0, 1.0),
            (1.0, 0.0, 0.0)),
 }
SNOM_amplitude = LinearSegmentedColormap('SNOM_amplitude', cmap_snom_amplitude)
plt.register_cmap(cmap=SNOM_amplitude)

cmap_snom_phase = {'red':   ((0.0, 0.0, 0.0),
           (0.33, 0.0, 0.0),
           (0.66, 1.0, 1.0),
           (1.0, 1.0, 0.0)),

  'green': ((0.0, 0.0, 0.0),
           (0.33, 0.0, 0.0),
           (0.66, 1.0, 1.0),
           (1.0, 0.0, 0.0)),

  'blue':  ((0.0, 0.0, 0.0),
           (0.33, 1.0, 1.0),
           (0.66, 1.0, 1.0),
           (1.0, 0.0, 0.0)),
 }
SNOM_phase = LinearSegmentedColormap('SNOM_phase', cmap_snom_phase)
plt.register_cmap(cmap=SNOM_phase)

cmap_snom_realpart = {
  'red':   ((0.0, 0.0, 0.0),
           (0.5, 1.0, 1.0),
           (1.0, 1.0, 0.0)),

  'green': ((0.0, 0.0, 0.0),
           (0.5, 1.0, 1.0),
           (1.0, 0.0, 0.0)),

  'blue':  ((0.0, 1.0, 1.0),
           (0.5, 1.0, 1.0),
           (1.0, 0.0, 0.0)),
}
SNOM_realpart = LinearSegmentedColormap('SNOM_realpart', cmap_snom_realpart)
plt.register_cmap(cmap=SNOM_realpart)
"""

# replace old maps with perceptual colormaps
SNOM_amplitude = cc.cm.CET_L3
# SNOM_phase = cc.cm.CET_C3s
SNOM_phase = cc.cm.CET_C3
# SNOM_realpart = cc.cm.CET_D1A 
SNOM_realpart = cc.cm.CET_D1
# SNOM_realpart = cc.cm.CET_D9
SNOM_height = cc.cm.CET_L2

# all implemented colormaps
all_colormaps = {
    "<SNOM_amplitude>": SNOM_amplitude,
    "<SNOM_height>": SNOM_height,
    "<SNOM_phase>": SNOM_phase,
    "<SNOM_realpart>": SNOM_realpart
}



"""Some additional functions for the snom_analysis package."""

##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import numpy as np

def set_nan_to_zero(data) -> np.array:
    xres = len(data[0])
    yres = len(data)
    for y in range(yres):
        for x in range(xres):
            if str(data[y][x]) == 'nan':
                data[y][x] = 0
    return data       

# needed for the realign function
def gauss_function(x, A, mu, sigma, offset):
    return A*np.exp(-(x-mu)**2/(2.*sigma**2)) + offset

def get_largest_abs(val1, val2):
    if abs(val1) > abs(val2): return abs(val1)
    else: return abs(val2)

def calculate_colorbar_size(fig, ax, colorbar_size):
    """This function converts a colorbar size in % of the fig width to a colorbar size in % of the axis width."""
    # size of the figure in inches
    fig_width = fig.get_figwidth()
    # size of the axis in inches
    ax_width = ax.get_position().width * fig_width
    # calculate the size of the colorbar in percent
    # it should always be x % of the figure width
    return colorbar_size * fig_width / ax_width 

def mean_index_array(array, interpolation=4):
    """Returns the index of the mean in the given list. The mean is calculated by interpolating the data and then calculating the mean index of the half area.

    Args:
        array (np.array or list): data array to calculate the mean index from
        interpolation (int, optional): how many datapoints to interpolate in between the given data. Defaults to 4.

    Returns:
        float: the mean index of the data
    """
    from scipy.interpolate import interp1d
    # create a spline interpolation of the data with x times the resolution
    x = np.arange(len(array))
    xn = np.linspace(0, len(array)-1, len(array)*interpolation)
    spline = interp1d(x, array, kind='cubic')
    array_interp = spline(xn)
    # calculate the mean index
    mean = 0
    half_area = np.sum(array_interp)/2
    lower_sum = 0
    for i in range(len(array_interp)):
        if lower_sum + array_interp[i] <= half_area:
            lower_sum += array_interp[i]
            mean = i
        else:
            break
    return mean/interpolation 


##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import tkinter as tk
import numpy as np
from PIL import Image, ImageTk
# import cv2 # not needed anymore, and very big package

#testing
import matplotlib.pyplot as plt
from .snom_colormaps import SNOM_amplitude, SNOM_phase, SNOM_height


def select_data_range(data, channel):
    root = tk.Tk()
    selector = ArraySelector(root, data, channel)
    root.mainloop()

    # selection = selector.selection
    start = selector.start
    end = selector.end
    is_horizontal = selector.is_horizontal
    inverted = selector.inverted
    return start, end, is_horizontal, inverted

class ArraySelector:
    def __init__(self, root, data, channel):
        self.root = root
        self.array = data
        self.original_array = self.array.copy()
        self.array = ((self.array - np.min(self.array)) / (np.max(self.array) - np.min(self.array)) * 255).astype(np.uint8)
        self.channel = channel
        self.height, self.width = data.shape
        self.original_height, self.original_width = self.original_array.shape

        # additional parameters
        self.highlighted_image = None
        self.rect = None
        self.start = None
        self.end = None
        self.dragging = False
        self.resizing = None  # 'left', 'right', or None
        self.inverted = False  # Track if the selection is inverted
        self.is_horizontal = True  # Track selection mode
        
        # scale the data to a size usable for the canvas, min should be 300x300 pixels
        # if both axes are smaller than 300 pixels then scale the data such that the bigger axis has at least 300 pixels
        min_size = 400
        max_size = 800
        self.scaling_factor = 1
        # first check if the data is already larger than the minimum size in at least one axis
        if self.width > min_size or self.height > min_size:
            pass
        # if both axes are smaller than the minimum size then scale the data such that the bigger axis has at least min_size pixels
        elif self.width < min_size and self.height < min_size:
            if self.width > self.height:
                self.scaling_factor = min_size/self.width
            else:
                self.scaling_factor = min_size/self.height
        elif self.width < min_size:
            self.scaling_factor = min_size/self.width
        elif self.height < min_size:
            self.scaling_factor = min_size/self.height
        # check if one of the axes is larger than the maximum size
        if self.width > max_size or self.height > max_size:
            if self.width > self.height:
                self.scaling_factor = max_size/self.width
            else:
                self.scaling_factor = max_size/self.height
        elif self.width > max_size:
            self.scaling_factor = max_size/self.width
        elif self.height > max_size:
            self.scaling_factor = max_size/self.height
        self.update_scaling_factor()

        # change colormap depending on the channel
        if ('Z' in self.channel) or ('MT' in self.channel):
            self.cmap = SNOM_height
        elif ('P' or 'arg') in self.channel:
            self.cmap = SNOM_phase
        elif ('A' or 'abs') in self.channel:
            self.cmap = SNOM_amplitude
        elif ('H' or 'height') in self.channel:
            self.cmap = SNOM_height
        else:
            self.cmap = 'gray'
            print('Unknown channel, could not find the proper colormap!')

        canvas_width = self.width*self.scaling_factor
        canvas_height = self.height*self.scaling_factor

        self.canvas = tk.Canvas(self.root, width=canvas_width, height=canvas_height)

        self.canvas.grid(row=0, column=0, sticky='nsew')

        self.fill_canvas()

        self.canvas.bind("<ButtonPress-1>", self.on_button_press)
        self.canvas.bind("<B1-Motion>", self.on_mouse_drag)
        self.canvas.bind("<ButtonRelease-1>", self.on_button_release)
        self.canvas.bind("<Motion>", self.on_mouse_move)
        self.canvas.bind("<Configure>", self.on_windowsize_changed)
        
        self.toggle_button = tk.Button(root, text="Toggle Selection Mode", command=self.toggle_selection_mode)
        self.toggle_button.grid(row=1, column=0)
        
        self.invert_button = tk.Button(root, text="Invert Selection", command=self.invert_selection)
        self.invert_button.grid(row=2, column=0)

        self.select_button = tk.Button(root, text="Get Coordinates", command=self.get_coordinates)
        self.select_button.grid(row=3, column=0)
        
        # Center the window
        root.eval(f'tk::PlaceWindow {str(self.root)} center')

        # configure canvas to scale with window
        self.root.grid_rowconfigure(0, weight=1)
        self.root.grid_columnconfigure(0, weight=1)

    def update_scaling_factor(self):
        if self.scaling_factor != 1:
            if self.scaling_factor > 1:
                self.scaling_factor = round(self.scaling_factor) # allow for scaling up to an arbitrary integer of the original size
            else:
                self.scaling_factor = 0.5 # allow for scaling down to half the size

    def on_button_press(self, event):
        if self.resizing:
            self.dragging = True
            return
        
        if self.is_horizontal:
            self.start = event.x
            self.end = event.x
            if self.rect:
                self.canvas.delete(self.rect)
            self.rect = self.canvas.create_rectangle(self.start, 0, self.start, self.canvas.winfo_height(), outline='red', width=2)
        else:
            self.start = event.y
            self.end = event.y
            if self.rect:
                self.canvas.delete(self.rect)
            self.rect = self.canvas.create_rectangle(0, self.start, self.canvas.winfo_width(), self.start, outline='red', width=2)

    def on_mouse_drag(self, event):
        if self.dragging:
            if self.resizing == 'left':
                if self.is_horizontal:
                    self.start = max(0, event.x)
                else:
                    self.start = max(0, event.y)
            elif self.resizing == 'right':
                if self.is_horizontal:
                    self.end = min(self.width, event.x)
                else:
                    self.end = min(self.height, event.y)
        else:
            if self.is_horizontal:
                # self.end = event.x
                if event.x < 0:
                    self.end = 0
                elif event.x > self.width:
                    self.end = self.width
                else:
                    self.end = event.x
            # else:
            #     self.end = event.y
            else:
                self.end = event.y
                if event.y < 0:
                    self.start = 0
                elif event.y > self.height:
                    self.end = self.height

        # Prevent overlapping
        if self.start > self.end:
            self.start, self.end = self.end, self.start

        if self.is_horizontal:
            self.canvas.coords(self.rect, self.start, 0, self.end, self.canvas.winfo_height())
        else:
            self.canvas.coords(self.rect, 0, self.start, self.canvas.winfo_width(), self.end)

        #print start and end
        print(f"Start: {self.start}, End: {self.end}")
        self.highlight_selection()

    def on_button_release(self, event):
        self.dragging = False
        self.resizing = None

    def on_mouse_move(self, event):
        if self.rect:
            if self.is_horizontal:
                left_border = self.start
                right_border = self.end
                border_width = 5
                
                if left_border - border_width < event.x < left_border + border_width:
                    self.canvas.config(cursor="sb_h_double_arrow")
                    self.resizing = 'left'
                elif right_border - border_width < event.x < right_border + border_width:
                    self.canvas.config(cursor="sb_h_double_arrow")
                    self.resizing = 'right'
                else:
                    self.canvas.config(cursor="")
                    self.resizing = None
            else:
                top_border = self.start
                bottom_border = self.end
                border_width = 5

                if top_border - border_width < event.y < top_border + border_width:
                    self.canvas.config(cursor="sb_v_double_arrow")
                    self.resizing = 'left'
                elif bottom_border - border_width < event.y < bottom_border + border_width:
                    self.canvas.config(cursor="sb_v_double_arrow")
                    self.resizing = 'right'
                else:
                    self.canvas.config(cursor="")
                    self.resizing = None

    def fill_canvas(self):#
        # self.array = cv2.resize(self.original_array, (int(self.scaling_factor*self.original_width), int(self.scaling_factor*self.original_height)),interpolation=cv2.INTER_NEAREST)
        # same using pillows image to get rid of cv2
        # self.array = Image.resize((int(self.scaling_factor*self.original_width), int(self.scaling_factor*self.original_height)), Image.NEAREST)
        # Normalize array to 0-255 for display
        # self.array = ((self.array - np.min(self.array)) / (np.max(self.array) - np.min(self.array)) * 255).astype(np.uint8)

        # self.height, self.width = self.array.shape
        # delete and redraw the image if already existing
        try:
            self.canvas.delete("all")
        except:
            pass
        # Normalize array to 0-255 for display
        # self.array = ((self.array - np.min(self.array)) / (np.max(self.array) - np.min(self.array)) * 255).astype(np.uint8)
        # self.image = Image.fromarray(np.uint8(self.cmap(self.original_array)*255))
        self.image = Image.fromarray(np.uint8(self.cmap(self.array)*255))
        # plt.imshow(self.image)
        # plt.show()
        # resize the image
        self.width = int(self.scaling_factor*self.original_width)
        # print(f'width: {self.width}')
        self.height = int(self.scaling_factor*self.original_height)
        # print(f'height: {self.height}')
        # print('resizing image')
        # print(f'scaling factor: {self.scaling_factor}')
        self.image = self.image.resize((self.width, self.height), Image.Resampling.LANCZOS)
        # plt.imshow(self.image)
        # plt.show()
        self.tk_image = ImageTk.PhotoImage(self.image)
        # self.canvas.create_image(self.width/2, self.height/2, image=self.tk_image)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=self.tk_image)
        # self.highlight_selection()

    def on_windowsize_changed(self, event):
        # if the canvas size becomse so large that the image scaled by an integer factor is smaller than the canvas size, then we need to rescale the image by that factor
        # get the current size of the canvas
        canvas_width = self.canvas.winfo_width()
        canvas_height = self.canvas.winfo_height()
        # get the current size of the data
        # height, width = self.array.shape
        height, width = self.image.size
        # if the current scaling factor is 0.5 we need to increment it by 0.5 not 1 
        # if the current scaling factor is 1 we need to increment it by 0.5 not 1
        if self.scaling_factor == 0.5 or self.scaling_factor == 1:
            scaling_increment = 0.5
        else:
            scaling_increment = 1
        # resize the image to the canvas size if the canvas size is larger than the image times the scaling factor, where the scaling factor is an integer
        if canvas_width > self.original_width*(self.scaling_factor+scaling_increment) and canvas_height > self.original_height*(self.scaling_factor+scaling_increment):
            # increment the scaling factor
            self.scaling_factor += scaling_increment
            self.update_scaling_factor()
            self.fill_canvas()
        elif canvas_width < width or canvas_height < height:
            # decrement the scaling factor
            self.scaling_factor -= scaling_increment
            self.update_scaling_factor()
            self.fill_canvas()        

    def highlight_selection(self):
        if self.highlighted_image:
            self.canvas.delete(self.highlighted_image)

        # Create an image for the highlighting based on inverted state
        # highlighted_img = np.zeros_like(self.array)
        # highlighted_img = np.ones_like(self.array).astype(np.uint8)*255
        highlighted_img = np.ones((self.height, self.width)).astype(np.uint8)*255
        # highlighted_img = self.array.copy()
        if self.inverted:
            highlighted_img[:, :] = 128  # Copy original array
            if self.is_horizontal:
                highlighted_img[:, self.start:self.end] = 255  # Set selected area to black
            else:
                highlighted_img[self.start:self.end, :] = 255  # Set selected area to black
        else:
            if self.is_horizontal and self.start is not None and self.end is not None:
                highlighted_img[:, self.start:self.end] = 128#self.array[:, self.start:self.end]  # Highlight selected area
            elif not self.is_horizontal and self.start is not None and self.end is not None:
                highlighted_img[self.start:self.end, :] = 128#self.array[self.start:self.end, :]  # Highlight selected area
        mask = Image.fromarray(highlighted_img)
        # print(f'mask size: {mask.size}')
        # print(f'original image size: {self.image.size}')
        # create an overlay in red with 30% opacity of the highlighted area with the original image
        overlay = Image.new('RGBA', self.image.size, (255, 0, 0, 0))  # Red with 30% opacity
        combined = Image.composite(self.image.convert('RGBA'), overlay, mask)
        # convert to rgb
        # combined = combined.convert('RGB')
        # now update the image
        self.tk_image = ImageTk.PhotoImage(combined)
        self.canvas.create_image(0, 0, anchor=tk.NW, image=self.tk_image)

    def invert_selection(self):
        self.inverted = not self.inverted  # Toggle inverted state
        self.highlight_selection()  # Update highlighting

    def toggle_selection_mode(self):
        # Clear current selection
        if self.rect:
            self.canvas.delete(self.rect)
            self.rect = None
        self.start = None
        self.end = None

        self.is_horizontal = not self.is_horizontal  # Toggle selection mode
        self.highlight_selection()  # Update highlighting for the new mode

    def get_coordinates(self):
        # reduze the coordinates to the original size
        if self.scaling_factor != 1:
            self.start = int(self.start/self.scaling_factor)
            self.end = int(self.end/self.scaling_factor)-1 # '-1' due to conversion between pixel in image and array index
        if self.start is not None and self.end is not None:
            if self.is_horizontal:
                print(f"Selected horizontal coordinates: {self.start}, {self.end}")
            else:
                print(f"Selected vertical coordinates: {self.start}, {self.end}")

        # close the window
        self.root.destroy()



##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import numpy as np
from matplotlib import pyplot as plt



def calculate_squared_mean_deviation(array_1, array_2) -> float:
    # check if both arrays are of equal length
    if len(array_1) != len(array_2):
        print('The lengths of the arrays must be equal!')
        return 0
    mean_deviation = 0
    N = len(array_1)
    for i in range(N):
        mean_deviation += (array_1[i]**2-array_2[i]**2)/N
    return mean_deviation

def calculate_squared_deviation(array_1, array_2) -> list:
    # check if both arrays are of equal length
    if len(array_1) != len(array_2):
        print('The lengths of the arrays must be equal!')
        return 0
    mean_deviation = []
    N = len(array_1)
    for i in range(N):
        # ignore '0' values
        if array_1[i] == 0 or array_2[i] == 0:
            mean_deviation.append(0)
        else:
            mean_deviation.append(abs(array_1[i]-array_2[i]))
    return mean_deviation

def _shift_array_1d_by_index(array_1, array_2, index) -> tuple:
    """This function shifts the second array with respect to the first array. The first array is modified to keep the same length.

    Args:
        array_1 (np.array): first and reference array
        array_2 (np.array): second array
        index (int): how much should be shifted, to the left if negative
    """
    # print(f'in shift 1d, index: {index}')
    if index > 0:
        array_1_new = array_1
        array_2_new = array_2
        for i in range(abs(index)):
            # print(f'apply shift, i: {i}')
            array_1_new = np.append(array_1_new, 0)
            array_2_new = np.insert(array_2_new, 0, 0)
    elif index < 0:
        array_1_new = array_1
        array_2_new = array_2
        for i in range(abs(index)):
            array_1_new = np.insert(array_1_new, 0, 0)
            array_2_new = np.append(array_2_new, 0)
    else:
        array_1_new = array_1
        array_2_new = array_2
    return array_1_new, array_2_new

def shift_array_2d_by_index(array_1, array_2, index) -> tuple:
    """This function shifts the second 2d array relative to the first. Zeroes will be created on the outside.

    Args:
        array_1 (np.ndarray): array1
        array_2 (np.ndarray): array2
        index (int): how much to shift, negative means to the left

    Returns:
        tuple: shifted arrays
    """
    x_res = len(array_1[0])
    y_res = len(array_1)
    # print(f'xres={x_res} yres={y_res}, index={index}')
    x_res_new = x_res + abs(index)
    #create new lists with modified resolution
    array_1_new = np.zeros((y_res, x_res_new))
    array_2_new = np.zeros((y_res, x_res_new))
    for y in range(y_res):
        test_1, test_2 = _shift_array_1d_by_index(array_1[y], array_2[y], index)
        # print(f'len test1: {len(test_1)}, len test 2: {len(test_2)}')
        array_1_new[y] = test_1
        array_2_new[y] = test_2
        # array_1_new[y], array_2_new[y] = _shift_array_1d_by_index(array_1[y], array_2[y], index)
    return array_1_new, array_2_new

def create_mean_array(array_1, array_2):
    #check if dimensions are identical
    if len(array_1) != len(array_2) and len(array_1[0]) != len(array_2[0]):
        print('The length of the given arrays is not identical.')
        exit()
    else:
        x_res = len(array_1[0])
        y_res = len(array_1)
        new_array = np.zeros((y_res, x_res))
        for y in range(y_res):
            for x in range(x_res):
                new_array[y][x] = (array_1[y][x] + array_2[y][x])/2
        return new_array

def create_mean_array_v2(array_1, array_2, index):
    """This variant is meant to keep the size of the original array!

    Args:
        array_1 (_type_): _description_
        array_2 (_type_): _description_

    Returns:
        _type_: _description_
    """
    #check if dimensions are identical
    if len(array_1) != len(array_2) and len(array_1[0]) != len(array_2[0]):
        print('The length of the given arrays is not identical.')
        exit()
    else:
        x_res = len(array_1[0])
        y_res = len(array_1)
        new_array = np.zeros((y_res, x_res))
        for y in range(y_res):
            if index > 0:
                for x in range(x_res - index):
                    new_array[y][x] = (array_1[y][x+index] + array_2[y][x])/2
            if index < 0:
                for x in range(x_res - abs(index)):
                    new_array[y][x] = (array_1[y][x] + array_2[y][x+abs(index)])/2
        return new_array

def minimize_deviation_1d(array_1, array_2, n_tries=5, display=True) -> int:
    """This function tries to find the optimal shift between two arrays to find the lowest deviation. Best to use leveled data.

    Args:
        array_1 (np.array): first array, typically height data
        array_2 (np.array): second array, typically height data
        n_tries (int, optional): the maximum shift expected for optimal overlap, will be applied symmetrically to right and left shift. Defaults to 5.

    Returns:
        int: the optimal shift index, left shift if negative
    """
    # try shifting to the right
    mean_deviations_right = []
    for i in range(n_tries):
        array_1_new, array_2_new = _shift_array_1d_by_index(array_1, array_2, i+1)
        mean_deviation_array = calculate_squared_deviation(array_1_new, array_2_new)
        mean_deviation = np.mean(mean_deviation_array)
        mean_deviations_right.append(mean_deviation)
        x = range(len(array_1_new))
        # plt.plot(x, array_1_new, label='array_2')
        # plt.plot(x, array_2_new, label='array_1')
        # plt.plot(x, mean_deviation_array, label="Mean deviation_array")
        # plt.hlines(mean_deviation, label="Mean deviation", xmin=0, xmax=len(array_1))
        # plt.legend()
        # plt.show()
    min_dev_right = min(mean_deviations_right)

    # try shifting to the left
    mean_deviations_left = []
    for i in range(n_tries):
        array_1_new, array_2_new = _shift_array_1d_by_index(array_1, array_2, -(i+1))
        mean_deviation_array = calculate_squared_deviation(array_1_new, array_2_new)
        mean_deviation = np.mean(mean_deviation_array)
        mean_deviations_left.append(mean_deviation)
        x = range(len(array_1_new))
        # plt.plot(x, array_1_new, label='array_2')
        # plt.plot(x, array_2_new, label='array_1')
        # plt.plot(x, mean_deviation_array, label="Mean deviation_array")
        # plt.hlines(mean_deviation, label="Mean deviation", xmin=0, xmax=len(array_1))
        # plt.legend()
        # plt.show()
    min_dev_left = min(mean_deviations_left)

    # try not shifting at all
    mean_deviation_array = calculate_squared_deviation(array_1, array_2)
    min_dev_center = np.mean(mean_deviation_array)
    
    min_dev_index = 0
    if min_dev_left <= min_dev_right and min_dev_left < min_dev_center:
        # print(f'The minimal deviation of {min_dev_left} is reached for a shift of {mean_deviations_left.index(min_dev_left)+1} to the left')
        array_1_new, array_2_new = _shift_array_1d_by_index(array_1, array_2, -(mean_deviations_left.index(min_dev_left)+1))
        min_dev_index = -(mean_deviations_left.index(min_dev_left)+1)
    elif min_dev_left > min_dev_right and min_dev_right < min_dev_center:
        # print(f'The minimal deviation of {min_dev_right} is reached for a shift of {mean_deviations_right.index(min_dev_right)+1} to the right')
        array_1_new, array_2_new = _shift_array_1d_by_index(array_1, array_2, +(mean_deviations_right.index(min_dev_right)+1))
        min_dev_index = (mean_deviations_right.index(min_dev_right)+1)
    elif min_dev_center <= min_dev_left and min_dev_center <= min_dev_right:
        array_1_new = array_1
        array_2_new = array_2
        min_dev_index = 0
    else:
        print(f'min_dev_left: {min_dev_left}, min_dev_center: {min_dev_center}, min_dev_right: {min_dev_right}')
        print('Oops, something went wrong, could not do the optimization')
        
    
    if display == True:
        # redo the best deviation:
        x = range(len(array_1_new))
        plt.plot(x, array_1_new, label='array_1')
        plt.plot(x, array_2_new, label='array_2')
        mean_deviation_array = calculate_squared_deviation(array_1_new, array_2_new)
        mean_deviation = np.mean(mean_deviation_array)
        plt.plot(x, mean_deviation_array, label="Mean deviation_array")
        plt.hlines(mean_deviation, label="Mean deviation", xmin=0, xmax=len(array_1_new))
        plt.legend()
        plt.show()

    return min_dev_index

def minimize_deviation_2d(array_1, array_2, n_tries=5, display=True) -> int:
    """This function tries to find the optimal shift between two arrays to find the lowest deviation. Best to use leveled data.

    Args:
        array_1 (np.array): first array, typically height data
        array_2 (np.array): second array, typically height data
        n_tries (int, optional): the maximum shift expected for optimal overlap, will be applied symmetrically to right and left shift. Defaults to 5.

    Returns:
        int: the optimal shift index, left shift if negative
    """
    y_shifts = [] # contains the optimal shifts per y data line
    rows = len(array_1)
    for i in range(rows):
        y_shifts.append(minimize_deviation_1d(array_1[i], array_2[i], n_tries, display))
    # plt.plot(range(rows), y_shifts, label="optimal dev index per row")
    # plt.legend(loc="upper left")
    # plt.show()
    return int(round(np.mean(y_shifts)))

def _shift_data(data, y_shifts):
    min_shift = min(y_shifts)
    max_shift = max(y_shifts)
    # print('shifts: ', y_shifts)
    if min_shift < 0:
        min_shift = abs(min_shift)
    else:
        min_shift = 0
    if max_shift > 0:
        max_shift = max_shift
    else:
        max_shift = 0
    # min and max shift now correspond to the empty spaces which have to be added in the beginning(min) and end(max) of the first line
    total_shift = min_shift +  max_shift
    if total_shift != 0:
        y_res = len(data)
        x_res = len(data[0])
        x_res_new = x_res + min_shift + max_shift
        data_new = np.zeros((y_res, x_res_new))
        for i in range(len(y_shifts)):
            row_data = data[i]
            index = y_shifts[i]
            num_of_inserts = min_shift + index
            num_of_appends = max_shift - index
            if num_of_inserts > 0:
                for j in range(num_of_inserts):
                    row_data = np.insert(row_data, 0, 0)
            if num_of_appends > 0:
                for j in range(num_of_appends):
                    row_data = np.append(row_data, 0)
            data_new[i] = row_data
    else:
        return data
    '''
    max_shift = abs(max(y_shifts, key=abs))
    print('shifts: ', y_shifts)
    print('max shift: ', max_shift)
    y_res = len(data)
    x_res = len(data[0])
    x_res_new = x_res + 2*max_shift
    data_new = np.zeros((y_res, x_res_new))
    for i in range(len(y_shifts)):
        row_data = data[i]
        index = y_shifts[i]
        num_of_appends = max_shift - index
        num_of_inserts = max_shift + index
        if num_of_inserts > 0:
            for j in range(num_of_inserts):
                row_data = np.insert(row_data, 0, 0)
        if num_of_appends > 0:
            for j in range(num_of_appends):
                row_data = np.append(row_data, 0)
        data_new[i] = row_data
    '''    
    return data_new

def minimize_drift(data:np.array, n_tries=5, display=False) -> np.array:
    """This function tries to shift subsequent lines of a single measurement to minimize the deviation.
    The function will return the drift corrected data.

    Returns:
        np.array: the drift corrected data
    """
    rows = len(data)
    # cols = len(data[0])
    y_shifts = [0] # the first row will not be shifted, each subsequent shift is related to the first line
    for y in range(rows-1):
        y_shifts.append(minimize_deviation_1d(data[0], data[y+1], n_tries, display))
    return _shift_data(data, y_shifts)







##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button
from mpl_toolkits.axes_grid1 import make_axes_locatable
from .snom_colormaps import *



def get_phase_offset(preview_data):
    phase_shift = PhaseSlider(preview_data)
    return phase_shift.phase_shift

class PhaseSlider():
    def __init__(self, preview_data):
        self.data = preview_data
        initial_shift = 0 # initial phase shift
        self.previous_shift = 0
        self.phase_shift = 0
        # create the plot
        self.fig, axis = plt.subplots()
        self.plot = plt.pcolormesh(self.data, cmap=SNOM_phase, vmin=0, vmax=np.pi*2)
        axis.invert_yaxis()
        divider = make_axes_locatable(axis)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = plt.colorbar(self.plot, cax=cax)
        cbar.ax.get_yaxis().labelpad = 15
        label = 'Phase'
        title = 'Shift the phase'
        cbar.ax.set_ylabel(label, rotation=270)
        axis.set_title(title)
        axis.axis('scaled')


        # adjust the main plot to make room for the sliders
        plt.subplots_adjust(left=0.25, bottom=0.25)
        # Make a horizontal slider to control the frequency.
        ax_phaseshift = plt.axes([0.25, 0.1, 0.65, 0.03])
        self.phase_slider = Slider(
            ax=ax_phaseshift,
            label='Frequency [Hz]',
            valmin=0.0,
            valmax=np.pi*2,
            valinit=initial_shift,
        )
        # register the update function with each slider
        self.phase_slider.on_changed(self.update)

        # Create a `matplotlib.widgets.Button` to accept the current value and close the window.
        accept = plt.axes([0.8, 0.025, 0.1, 0.04])
        button = Button(accept, 'Accept', hovercolor='0.975')
        button.on_clicked(self.accept)
        plt.show()
    
    def accept(self, event):
        #close the plot, the recent phase shift is saved in self.phase_shift and can be accessed from outside
        plt.close()        
    
    def update(self, val):
        self.phase_shift = val 
        self.shift_phase()
        self.plot.set_array(self.data)
        self.fig.canvas.draw_idle()
    
    def shift_phase(self):
        yres = len(self.data)
        xres = len(self.data[0])
        for y in range(yres):
            for x in range(xres):
                self.data[y][x] = (self.data[y][x] + (self.phase_shift - self.previous_shift)) % (2*np.pi)
        self.previous_shift = self.phase_shift



##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import matplotlib.pyplot as plt
# import matplotlib
import numpy as np
import matplotlib.pyplot as plt
from skimage import feature
from matplotlib.widgets import Slider, Button
from mpl_toolkits.axes_grid1 import make_axes_locatable # for colorbar
import scipy.ndimage as ndimage


# initial values for the sliders
initial_threshold_low = 0.2
initial_threshold_high = 0.8
initial_sigma = 4

class EdgeDetection():
    """This class creates a GUI to adjust the parameters for the edge detection algorithm.
    The user can adjust the sigma value, the low and high threshold for the edge detection.
    The user can accept the values or exit the GUI.
    Uppon accepting the values the GUI is closed and the values are saved to the class instance variables.
    """
    def __init__(self, data):
        self.data = data
        self.iterations = None
        # create the gui
        self.create_gui()
    
    def create_gui(self):
        self.fig, axis = plt.subplots()
        axis.imshow(self.data, cmap='gray')
        self.edges = self.calculate_edges(initial_sigma, initial_threshold_low, initial_threshold_high)
        masked_data = np.ma.masked_where(self.edges != 1, self.edges)
        self.plot = axis.imshow(masked_data, interpolation='none', cmap='viridis', vmin=0, vmax=1)
        axis.invert_yaxis()
        # divider = make_axes_locatable(axis)
        # cax = divider.append_axes("right", size="5%", pad=0.05)
        # cbar = plt.colorbar(self.plot, cax=cax)
        # cbar.ax.get_yaxis().labelpad = 15
        # label = 'Height'
        title = 'Adjust the sliders to optimize the edge detection. Press "Accept" to save the settings.'
        # cbar.ax.set_ylabel(label, rotation=270)
        axis.set_title(title)
        axis.axis('scaled')

        # adjust the main plot to make room for the sliders
        plt.subplots_adjust(left=0.25, bottom=0.25)
        # Make a horizontal slider to control the frequency.
        ax_threshold_low = plt.axes([0.25, 0.1, 0.65, 0.03])
        self.slider_threshold_low = Slider(
            ax=ax_threshold_low,
            label='Threshold low',
            valmin=0.0,
            valmax=1,
            valinit=initial_threshold_low,
        )
        # register the update function with each slider
        self.slider_threshold_low.on_changed(self.update)

        ax_threshold_high = plt.axes([0.25, 0.15, 0.65, 0.03])
        self.slider_threshold_high = Slider(
            ax=ax_threshold_high,
            label='Threshold high',
            valmin=0.0,
            valmax=1,
            valinit=initial_threshold_high,
        )
        # register the update function with each slider
        self.slider_threshold_high.on_changed(self.update)

        ax_sigma = plt.axes([0.25, 0.05, 0.4, 0.03])
        self.slider_sigma = Slider(
            ax=ax_sigma,
            label='Sigma',
            valmin=0.0,
            valmax=10,
            valinit=initial_sigma,
        )
        # register the update function with each slider
        self.slider_sigma.on_changed(self.update)

        # Create a `matplotlib.widgets.Button` to accept the current value and close the window.
        accept = plt.axes([0.8, 0.025, 0.1, 0.04])
        button = Button(accept, 'Accept', hovercolor='0.975')
        button.on_clicked(self.accept)

        exit = plt.axes([0.7, 0.025, 0.1, 0.04])
        button_exit = Button(exit, 'Exit', hovercolor='0.975')
        button_exit.on_clicked(self.exit)

        # show the plot in full screen, important for high resolution data
        mng = plt.get_current_fig_manager()
        mng.full_screen_toggle()

        plt.show()
        
    def calculate_edges(self, sigma, threshold_low, threshold_high):
        edges = feature.canny(self.data, sigma=sigma, use_quantiles=True, low_threshold=threshold_low, high_threshold=threshold_high)
        return edges

    def accept(self, event):
        #close the plot, the recent values are saved to class instance variables and can be accessed from outside
        plt.close()   
        self.threshold_high = self.slider_threshold_high.val
        self.threshold_low = self.slider_threshold_low.val
        self.sigma = self.slider_sigma.val   

    def exit(self, event):
        #close the plot, the recent values are saved to class instance variables and can be accessed from outside
        plt.close()
        # set the values to the initial values
        self.threshold_high = initial_threshold_high
        self.threshold_low = initial_threshold_low
        self.sigma = initial_sigma
    
    def update(self, val):
        self.phase_shift = val 
        # calculate the new data
        self.edges = self.calculate_edges(self.slider_sigma.val, self.slider_threshold_low.val, self.slider_threshold_high.val)
        # self.edges = feature.canny(adjusted_height_data, sigma=self.slider_sigma.val, use_quantiles=True, low_threshold=self.slider_threshold_low.val, high_threshold=self.slider_threshold_high.val)
        
        # if resolution of data is too high compared to screen resolution then dilate the edges
        # the iterations have to be calculated only once
        if self.iterations is None:
            # get resolution of data
            yres, xres = self.data.shape
            # get resolution of screen
            xscreen, yscreen = plt.gcf().get_size_inches() * plt.gcf().dpi
            # if the resolution of the data is higher than the screen resolution then approximate the iterations for the dilation
            if xres > xscreen or yres > yscreen:
                xres_factor = int(xres/xscreen)
                yres_factor = int(yres/yscreen)
                self.iterations = max(xres_factor, yres_factor)
            else:
                self.iterations = 0
        # if the iterations are not 0 then dilate the edges
        if self.iterations != 0:
            self.edges = ndimage.binary_dilation(self.edges, iterations=self.iterations)

        # update the plot
        # mask the edges where the edges are not 1 to make onl the edges visible in the overlay with the height data
        masked_data = np.ma.masked_where(self.edges != 1, self.edges)
        self.plot.set_data(masked_data)
        # self.fig.canvas.draw_idle() # not shure what the difference is between draw_idle and draw
        plt.draw()


##############################################################################
# Copyright (C) 2020-2025 Hans-Joachim Schill

# This file is part of snom_analysis.

# snom_analysis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# snom_analysis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with snom_analysis.  If not, see <http://www.gnu.org/licenses/>.
##############################################################################

# import numpy as np
# from matplotlib import pyplot as plt



def horizontal_profile(array):
    xres = len(array[0])
    yres = len(array)
    print(f'xres: {xres}')
    print(f'yres: {yres}')
    profile = []
    for x in range(xres):
        mean = 0
        for y in range(yres):
            mean += array[y][x]/yres
        profile.append(mean)
    return profile
